package de.devloop.openid;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpRequest.BodyPublishers;
import java.net.http.HttpResponse;
import java.net.http.HttpResponse.BodyHandlers;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import com.google.gson.Gson;

public class OpenID {
    private static final String CLIENT_ID = "vP9xF2s1yy2n6sR05jV6dguyMeOvIxCg1GarV71O";
    private static final String CLIENT_SECRET = "PrwGSMcucxYPkOdrb23jddWqyn31vphrxCUu9MGdLTCUnbk0OJI5oWCvO0khVhcnJNDbJaKWxNMxaC4bJ92jy8bDjtG6oaWG37qhuLRPMh5DKluZxsCMmCvQ8f9ZQckZ";

    private static final String REDIRECT_URL = "http://localhost:8080/mavor/authenticate";

    private static final String OAUTH_AUTH_URL = "https://auth.devloop.de/application/o/authorize/";
    private static final String OAUTH_TOKEN_URL = "https://auth.devloop.de/application/o/token/";
    private static final String OAUTH_USERINFO_URL = "https://auth.devloop.de/application/o/userinfo/";

    public AuthenticationUrl getAuthenticationUrl() {
        String state = UUID.randomUUID().toString();
        String url = String.format("%s?response_type=code&client_id=%s&redirect_uri=%s&state=%s&scope=openid email", OAUTH_AUTH_URL, CLIENT_ID, REDIRECT_URL, state);
        return new AuthenticationUrl(url, state);
    }

    private URI getUriObject(String url) throws OpenIdRequestException {
        try {
            return new URI(url);
        } catch (URISyntaxException e) {
            throw new OpenIdRequestException(String.format("Invalid URL: '%s'", url), e);
        }
    }

    public Token requestToken(String code) throws OpenIdRequestException {
        URI tokenUrl = getUriObject(OAUTH_TOKEN_URL);
        HashMap<String, String> tokenParameter = new HashMap<>();
        tokenParameter.put("grant_type", "authorization_code");
        tokenParameter.put("client_id", CLIENT_ID);
        tokenParameter.put("client_secret", CLIENT_SECRET);
        tokenParameter.put("code", code);
        tokenParameter.put("redirect_uri", REDIRECT_URL);

        HttpRequest tokenRequest = HttpRequest.newBuilder()
            .uri(tokenUrl)
            .header("Content-Type", "application/x-www-form-urlencoded")
            .header("Accept", "application/json")
            .POST(BodyPublishers.ofString(getFormDataAsString(tokenParameter)))
            .build();
        HttpClient tokenClient = HttpClient.newHttpClient();
        HttpResponse<String> tokenResponse;
        try {
            tokenResponse = tokenClient.send(tokenRequest, BodyHandlers.ofString());
        } catch (IOException | InterruptedException e) {
            throw new OpenIdRequestException("Requesting access token failed", e);
        }

        Gson gson = new Gson();
        return gson.fromJson(tokenResponse.body(), Token.class);
    }

    public UserInfo requestUserInfo(Token token) throws OpenIdRequestException {
        URI userInfoUrl = getUriObject(OAUTH_USERINFO_URL);
        HttpRequest userInfoRequest = HttpRequest.newBuilder()
            .uri(userInfoUrl)
            .header("Accept", "application/json")
            .header("Authorization", "Bearer " + token.getAccessToken())
            .GET()
            .build();
        HttpClient userInfoClient = HttpClient.newHttpClient();
        HttpResponse<String> userInfoResponse;
        try {
            userInfoResponse = userInfoClient.send(userInfoRequest, BodyHandlers.ofString());
        } catch (IOException | InterruptedException e) {
            throw new OpenIdRequestException("Requesting user info failed", e);
        }

        Gson gson = new Gson();
        return gson.fromJson(userInfoResponse.body(), UserInfo.class);
    }

    private String getFormDataAsString(Map<String, String> formData) {
        StringBuilder formBodyBuilder = new StringBuilder();
        for (Map.Entry<String, String> singleEntry : formData.entrySet()) {
            if (formBodyBuilder.length() > 0) {
                formBodyBuilder.append("&");
            }
            formBodyBuilder.append(URLEncoder.encode(singleEntry.getKey(), StandardCharsets.UTF_8));
            formBodyBuilder.append("=");
            formBodyBuilder.append(URLEncoder.encode(singleEntry.getValue(), StandardCharsets.UTF_8));
        }
        return formBodyBuilder.toString();
    }
}
